# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for pipeline task.

This base class extends taskflow Task class, with additional methods and
properties used by the GAPIC pipeline."""

import logging
import subprocess

from gcloud import logging as cloud_logging

from taskflow.task import Task

from artman.utils.logger import logger as artman_logger
from artman.utils.logger import output_logger
from artman.utils.logger import OUTPUT


class TaskBase(Task):

    cloud_logger = None

    def __init__(self, *args, **kwargs):
        super(TaskBase, self).__init__(*args, **kwargs)

    def log(self, msg, logger=artman_logger, level=logging.INFO):
        """Do local logging, and optionally cloud logging.

        Args:
            msg (str): The message to be logged.
            logger (logging.Logger): The logger to use. This defaults to
                logging.getLogger('artman').
            level (int): The log level. Defaults to logging.INFO.
        """
        logger.log(level, msg)

    def exec_command(self, args):
        """ Execute command and return output.

        TODO(ethanbao): Use subprocess.Popen which is recommended."""
        try:
            self.log(' '.join(args), level=logging.DEBUG)
            output = subprocess.check_output(args, stderr=subprocess.STDOUT)
            if output:
                output = output.decode('utf8')
                self.log(output, logger=output_logger, level=OUTPUT)
            return output
        except subprocess.CalledProcessError as e:
            self.log(e.output.decode('utf8'), logger=output_logger,
                     level=logging.ERROR)
            raise e


class EmptyTask(TaskBase):
    """An empty task that can be used by languages when they do not need to
    implement some functionality.
    """
    def execute(self):
        pass

    def validate(self):
        return []



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Factory function that recreates pipeline based on pipeline name and
kwargs."""


from artman.pipelines import pipeline_base
# These are required to list the subclasses of pipeline_base
from artman.pipelines import sample_pipeline
from artman.pipelines import gapic_generation
from artman.pipelines import grpc_generation
from artman.pipelines import core_generation
from artman.utils.logger import logger

__all__ = (
    'core_generation', 'gapic_generation',
    'grpc_generation', 'make_pipeline', 'make_pipeline_flow',
    'pipeline_base', 'sample_pipeline',
)


def make_pipeline_flow(pipeline_name, **kwargs):
    """Factory function to make a GAPIC pipeline.

    Because the GAPIC pipeline is using OpenStack Taskflow, this factory
    function is expected to be a function (or staticmethod) which is
    reimportable (aka has a well defined name that can be located by the
    __import__ function in python, this excludes lambda style functions and
    instance methods). The factory function name will be saved into the
    logbook, and it will be imported and called to create the workflow objects
    (or recreate it if resumption happens).  This allows for the pipeline to be
    recreated if and when that is needed.

    """
    return make_pipeline(pipeline_name, **kwargs).flow


def make_pipeline(pipeline_name, **kwargs):
    for cls in _rec_subclasses(pipeline_base.PipelineBase):
        if cls.__name__ == pipeline_name:
            logger.info("Creating %s." % pipeline_name)
            return cls(**kwargs)
    raise ValueError("Invalid pipeline name: %s" % pipeline_name)


def _rec_subclasses(cls):
    """Returns all recursive subclasses of a given class (i.e., subclasses,
    sub-subclasses, etc.)"""
    subclasses = []
    if cls.__subclasses__():
        for subcls in cls.__subclasses__():
            subclasses.append(subcls)
            subclasses += _rec_subclasses(subcls)
    return subclasses



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Util class for programming-language-specific things.
"""

import os
import os.path


class LanguageParams:
    def code_root(self, base_dir):
        return base_dir


class JavaParams(LanguageParams):
    def code_root(self, base_dir):
        return os.path.join(base_dir, 'src', 'main', 'java')


class PhpParams(LanguageParams):
    def code_root(self, base_dir):
        return os.path.join(base_dir, 'src')


LANG_PARAMS_MAP = {
    'python': LanguageParams(),
    'ruby': LanguageParams(),
    'nodejs': LanguageParams(),
    'java': JavaParams(),
    'go': LanguageParams(),
    'csharp': LanguageParams(),
    'php': PhpParams(),
}




# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for code generation pipelines."""

import time
import uuid
import importlib

from artman.utils import task_utils
from artman.pipelines import pipeline_base
from artman.tasks import io_tasks
from taskflow.patterns import linear_flow


# kwargs required by multiple pipelines
COMMON_REQUIRED = ['src_proto_path', 'import_proto_path', 'toolkit_path', 'root_dir',
                   'output_dir', 'api_name', 'api_version',
                   'organization_name', 'proto_package']

COMMON_DISCO_REQUIRED = ['discovery_doc', 'toolkit_path', 'root_dir', 'output_dir',
                         'api_name', 'api_version', 'organization_name']


class CodeGenerationPipelineBase(pipeline_base.PipelineBase):
    """Base class for GAPIC, gRPC, and Core code generation pipelines, and
    for GAPIC config generation pipeline.
    """
    def __init__(self, task_factory, **kwargs):
        self.task_factory = task_factory
        super(CodeGenerationPipelineBase, self).__init__(
            **kwargs)

    def do_build_flow(self, **kwargs):
        tasks = task_utils.instantiate_tasks(
            [io_tasks.PrepareGoogleapisDirTask], kwargs)
        tasks += self.task_factory.get_tasks(**kwargs)
        flow = linear_flow.Flow('CodeGenerationPipeline')
        flow.add(*tasks)
        return flow

    def validate_kwargs(self, **kwargs):
        # TODO(garrettjones) fix required to be relative to pipeline.
        # Ideally this should just be computed dynamically based
        # on the pipeline's tasks.
        _validate_exists(self.task_factory.get_validate_kwargs(),
                                      **kwargs)
        _validate_does_not_exist(
            self.task_factory.get_invalid_kwargs(), **kwargs)


class TaskFactoryBase(object):
    """Abstract base class for task factories.

    Task factory objects are used by CodeGenerationPipelineBase.
    Subclasses must implement the get_tasks and get_validate_kwargs
    methods.
    """

    def get_tasks(self, **kwargs):
        """Abstract method, subclasses must implement this method and return
        a list of tasks.
        """
        raise NotImplementedError('Subclass must implement abstract method')

    def get_validate_kwargs(self):
        """Abstract method, subclasses must implement this method and return
        a list of required keyword arguments.
        """
        raise NotImplementedError('Subclass must implement abstract method')

    def get_invalid_kwargs(self):
        """Abstract method, subclasses must implement this method and return
        a list of required keyword arguments.
        """
        raise NotImplementedError('Subclass must implement abstract method')


def _validate_exists(required, **kwargs):
    for arg in required:
        if arg not in kwargs:
            raise ValueError('{0} must be provided'.format(arg))


def _validate_does_not_exist(unsupported, **kwargs):
    for arg in unsupported:
        if arg in kwargs:
            raise ValueError('{0} is not supported'.format(arg))



#!/usr/bin/env python

# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Setup tool for artman."""

import io
import os
import re
import setuptools

cur_dir = os.path.realpath(os.path.dirname(__file__))

# version is defined in Dockerfile: ENV ARTMAN_VERSION version
current_version = None
with io.open(os.path.join(cur_dir, 'Dockerfile')) as dockerfile:
    for line in dockerfile:
        match = re.match(r'^ENV ARTMAN_VERSION (\S+)$', line)
        if match:
            current_version = match.group(1)
            break
if not current_version:
    print('Cannot determine version from Dockerfile. Exiting.')
    exit(1)

with io.open(os.path.join(cur_dir, 'requirements.txt')) as requirements_file:
    requirements = requirements_file.read().strip().split('\n')

setuptools.setup(
    name='googleapis-artman',
    version=current_version,
    description='Google API artifact manager',
    author='Google LLC',
    author_email='googleapis-packages@google.com',
    url='https://github.com/googleapis/artman',
    license='Apache-2.0',
    install_requires=requirements,
    packages=setuptools.find_packages(),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'artman = artman.cli.main:main',
            'configure-artman = artman.cli.configure:configure',
            'start-artman-conductor = artman.cli.conductor:start',
        ],
    },
    scripts=[
        'test/smoketest_artman.py',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ]
)



# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for protoc tasks"""

import collections.abc
import io
import os
import re
import subprocess
import types
import sys

import six

from artman.utils import lang_params
from artman.utils import task_utils
from artman.utils.logger import logger


class _SimpleProtoParams(object):
    def __init__(self, language):
        self.language = language
        self.path = None
        self.params = lang_params.LANG_PARAMS_MAP[language]

    def code_root(self, output_dir):
        return self.params.code_root(output_dir)

    def proto_plugin_path(self, plugin_args=None):
        return None

    def plugin_out_param(self, output_dir, plugin_args=None):
        return None

    def default_lang_out_value(self, with_grpc):
        return '{root}'

    def lang_out_param(self, output_dir, with_grpc, language_out_override):
        parameter_key = '--{language}_out'.format(language=self.language)
        parameter_value_template = language_out_override or self.default_lang_out_value(with_grpc)
        parameter_value = parameter_value_template.format(root=self.code_root(output_dir))
        return '{}={}'.format(parameter_key, parameter_value)

    def grpc_plugin_path(self, dummy_toolkit_path):
        if self.path is None:
            self.path = subprocess.check_output(
                ['which', 'grpc_{}_plugin'.format(self.language)],
                stderr=subprocess.STDOUT).decode('utf-8')
            self.path = six.text_type(self.path)[:-1]
        return self.path

    def grpc_out_param(self, output_dir):
        return '--grpc_out=' + self.code_root(output_dir)

    @property
    def proto_compiler_command(self):
        return [protoc_binary_name(self.language)]


class _JavaProtoParams(_SimpleProtoParams):
    def __init__(self):
        super(_JavaProtoParams, self).__init__('java')

    def code_root(self, output_dir):
        return self.params.code_root(output_dir)

    def proto_plugin_path(self):
        return subprocess.check_output(
            ['which', 'gapic_plugin.py'],
            stderr=subprocess.STDOUT).strip().decode('utf-8')

    def plugin_out_param(self, output_dir, plugin_args=None):
        # Java proto plugin requires the gapic yaml as a plugin arg
        if plugin_args:
            return '--plgn_out={}:{}'.format(plugin_args,
                                             self.code_root(output_dir))
        else:
            return None

    def grpc_plugin_path(self, toolkit_path):
        return task_utils.get_java_tool_path(toolkit_path, 'protoGenGrpcJavaExe')

    def grpc_out_param(self, output_dir):
        return '--grpc_out=' + self.code_root(output_dir)

    @property
    def proto_compiler_command(self):
        return [protoc_binary_name('java')]


class _GoProtoParams(_SimpleProtoParams):
    def __init__(self):
        super(_GoProtoParams, self).__init__('go')

    def code_root(self, output_dir):
        return self.params.code_root(output_dir)

    def default_lang_out_value(self, with_grpc):
        value = '{root}'
        if with_grpc:
            value = 'plugins=grpc:' + value
        return value

    def grpc_plugin_path(self, toolkit_path):
        # Go gRPC code is generated through --go_out=plugin=grpc, no grpc
        # specific plugin.
        return None

    def grpc_out_param(self, output_dir):
        # Go gRPC output directory is specified from --go_out, thus this
        # returns None.
        return None

    @property
    def proto_compiler_command(self):
        return [protoc_binary_name('go')]


class _PhpProtoParams(_SimpleProtoParams):
    def __init__(self):
        super(_PhpProtoParams, self).__init__('php')

    def code_root(self, output_dir):
        return self.params.code_root(output_dir)

    def grpc_out_param(self, output_dir):
        return '--grpc_out={}:{}'.format(
            'class_suffix=GrpcClient',
            self.code_root(output_dir)
        )

    @property
    def proto_compiler_command(self):
        return [protoc_binary_name('php')]


class _RubyProtoParams(_SimpleProtoParams):
    def __init__(self):
        super(_RubyProtoParams, self).__init__('ruby')

    def code_root(self, output_dir):
        return self.params.code_root(output_dir)

    def grpc_plugin_path(self, dummy_toolkit_path):
        # No plugin for grpc_toos_ruby_protoc
        return None

    def grpc_out_param(self, output_dir):
        return '--grpc_out=' + self.code_root(output_dir)

    @property
    def proto_compiler_command(self):
        return ['grpc_tools_ruby_protoc']


class _PythonProtoParams(_SimpleProtoParams):
    def __init__(self):
        super(_PythonProtoParams, self).__init__('python')

    def code_root(self, output_dir):
        return self.params.code_root(output_dir)

    def lang_out_param(self, output_dir, with_grpc, language_out_override):
        parameter_key = '--{language}_out'.format(language=self.language)
        parameter_value_template = language_out_override or self.default_lang_out_value(with_grpc)
        parameter_value = parameter_value_template.format(root=self.code_root(output_dir))
        # appending --pydocstring_out here is ugly, but it's not a good time to refactor now
        # since this tool is going to be discontinued really soon.
        append_value = '--pydocstring_out={root}'.format(root=self.code_root(output_dir))
        return '{}={} {}'.format(parameter_key, parameter_value, append_value)

    def grpc_plugin_path(self, dummy_toolkit_path):
        # No plugin for grpc.tools
        return None

    def grpc_out_param(self, output_dir):
        return '--grpc_python_out=' + self.code_root(output_dir)

    @property
    def proto_compiler_command(self):
        return [sys.executable, '-m', 'grpc_tools.protoc']


def protoc_binary_name(language):
    language = language.lower()
    top_dir = os.path.realpath(os.path.dirname(__file__ + "/../../../"))
    protoc_install_path = os.path.join(top_dir, 'install_protoc.sh')

    if not os.path.exists(protoc_install_path):
      # no script in its default location: we're likely running locally and not in Docker image
      return 'protoc'

    with io.open(protoc_install_path) as protoc_install_file:
        for line in protoc_install_file:
            match = re.match(r'^protobuf_versions\[' + language.lower() + r'\]=(\S+)$', line)
            if match:
                current_version = match.group(1)
                break
    if not current_version:
        raise IOError('Cannot determine version from Dockerfile. Using default version %s.' % current_version)
    return 'protoc-' + current_version


PROTO_PARAMS_MAP = {
    'ruby': _RubyProtoParams(),
    'java': _JavaProtoParams(),
    'go': _GoProtoParams(),
    'csharp': _SimpleProtoParams('csharp'),
    'php': _PhpProtoParams(),
    'python': _PythonProtoParams(),
    'nodejs': _SimpleProtoParams('nodejs'),
}


def group_by_go_package(proto_files):
    """Groups the file paths by `option go_package` in the file.
    This reflects the logic in https://github.com/google/go-genproto/blob/master/regen.go

    Returns:
        A dict mapping go_package to the list of proto files in the package.
    """

    def go_pkg(file):
        prefix = 'option go_package ='
        with open(file, encoding='utf-8') as f:
            for line in f:
                if line.startswith(prefix):
                    # syntax is: option go_package = "path/to/package;nickname";
                    # the nickname is optional
                    line = line[len(prefix):]
                    line = line.strip().strip('";')
                    # find the nickname
                    semi = line.find(';')
                    if semi >= 0:
                        line = line[:semi]
                    return line
        return ""

    pkgs = {}
    for file in proto_files:
        pkg = go_pkg(file)
        pkgs.setdefault(pkg, []).append(file)

    return pkgs


def protoc_header_params(proto_path,
                          toolkit_path):
    proto_path = proto_path[:]
    proto_path.append(_find_protobuf_path(toolkit_path))
    return (['--experimental_allow_proto3_optional'] + ['--proto_path=' + path for path in proto_path])


def protoc_desc_params(output_dir, desc_out_file):
    return (['--include_imports',
             '--include_source_info',
             '-o', os.path.join(output_dir, desc_out_file)])


def protoc_proto_params(proto_params, pkg_dir, gapic_yaml, with_grpc, language_out_override):
    params = []
    lang_param = proto_params.lang_out_param(pkg_dir, with_grpc, language_out_override)
    if lang_param:
        params += lang_param.split(' ')
    return params


def protoc_grpc_params(proto_params, pkg_dir, toolkit_path):
    params = []
    plugin_param = proto_params.grpc_plugin_path(toolkit_path)
    if plugin_param:
        params.append('--plugin=protoc-gen-grpc={}'.format(plugin_param))
    grpc_param = proto_params.grpc_out_param(pkg_dir)
    if grpc_param:
        params.append(grpc_param)
    return params


def protoc_plugin_params(proto_params, pkg_dir, gapic_yaml):
    params = []
    plugin_param = proto_params.proto_plugin_path()
    plugin_out = proto_params.plugin_out_param(pkg_dir, gapic_yaml)
    if plugin_param and plugin_out:
        params.append('--plugin=protoc-gen-plgn={}'.format(plugin_param))
        params.append(plugin_out)
    return params


def protoc_common_resources_params(root_dir, common_resources=None):
    resources = common_resources
    if resources is None:
        default = os.path.join("google", "cloud", "common_resources.proto")
        resources = [default]

    common_resources_includes = []
    common_resources_paths = []
    for res in resources:
        path = os.path.join(root_dir, res)
        if os.path.exists(path):
            common_resources_paths.append(path)
            common_resources_includes.append("-I{}={}".format(res, path))
        else:
            logger.warn("Could not find resources file at: {}".format(path))
    return (common_resources_includes, common_resources_paths)


def find_google_dir_index(src_proto_path):
    matches = list(re.finditer('(?:\\A|[/\\\\])(google|grafeas)(?=\\Z|[/\\\\])',
                               src_proto_path))
    if len(matches) == 0:
        raise ValueError('src_proto_path did not contain "google" '
                         'in path as expected. src_proto_path: '
                         '"{}"'.format(src_proto_path))
    return matches[-1].start(1)


def pkg_root_dir(output_dir, api_name, api_version, organization_name,
                 language, prefix=None):
    pkg_name = task_utils.api_full_name(
        api_name, api_version, organization_name)
    if prefix is not None:
        pkg_name = prefix + pkg_name
    return os.path.join(output_dir, language, pkg_name)


def prepare_proto_pkg_dir(output_dir, api_name, api_version, organization_name,
                          language):
    return prepare_pkg_dir(output_dir, api_name, api_version, organization_name,
                            language, "proto-")


def prepare_grpc_pkg_dir(output_dir, api_name, api_version, organization_name,
                         language):
    return prepare_pkg_dir(output_dir, api_name, api_version, organization_name,
                            language, "grpc-")


def prepare_pkg_dir(output_dir, api_name, api_version, organization_name,
                    language, prefix):
    proto_params = PROTO_PARAMS_MAP[language]
    pkg_dir = pkg_root_dir(
        output_dir, api_name, api_version, organization_name, language, prefix)
    subprocess.check_output([
        'mkdir', '-p', proto_params.code_root(pkg_dir)],
        stderr=subprocess.STDOUT)
    return pkg_dir


def find_protos(proto_paths, excluded_proto_path):
    """Searches along `proto_paths` for .proto files and returns a generator of
    paths"""
    if not isinstance(proto_paths, (types.GeneratorType, collections.abc.MutableSequence)):
        raise ValueError("proto_paths must be a list")
    for path in proto_paths:
        if os.path.isdir(path):
            for root, _, files in os.walk(path):
                for proto in files:
                    is_excluded = _is_proto_excluded(os.path.join(root, proto),
                                                     excluded_proto_path)
                    if os.path.splitext(proto)[1] == '.proto' and not is_excluded:
                        yield os.path.join(root, proto)
        elif os.path.isfile(path) and os.path.splitext(path)[1] == '.proto':
            yield path


def list_files_recursive(path):
    for root, _, files in os.walk(path):
        for f in files:
            yield os.path.join(root, f)


_php_replacements = [
    ('\\Google\\Protobuf\\Empty', '\\Google\\Protobuf\\GPBEmpty'),
]
def php_proto_rename(contents):
    for src, target in _php_replacements:
        contents = contents.replace(src, target)
    return contents


def _is_proto_excluded(proto, excluded_proto_path):
    for excluded_path in excluded_proto_path:
        if excluded_path in proto:
            return True
    return False

_protobuf_path = None
def _find_protobuf_path(toolkit_path):
    """Fetch and locate protobuf source"""
    global _protobuf_path
    if not _protobuf_path:
        logger.debug('Searching for latest protobuf source')
        _protobuf_path = task_utils.get_java_tool_path(toolkit_path, 'protobufJavaDir')
    return _protobuf_path



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Artman config reader that parses, validates and normalizes the passed artman
config.
"""

from __future__ import absolute_import
import io
import json
import os

from google.protobuf import json_format
import yaml

from artman.config.proto.config_pb2 import Artifact, Config
from artman.config.proto.user_config_pb2 import UserConfig
from artman.utils.logger import logger

# Error messages
CONFIG_NOT_FOUND_ERROR_MESSAGE_FORMAT = (
    'Artman YAML cannot be found at `%s`. Please check the file location.')
# TODO(ethanbao): Add a reference link once Artman schema spec is publicly
# available.
INVALID_CONFIG_ERROR_MESSAGE_FORMAT = 'Artman YAML %s is invalid.'
INVALID_USER_CONFIG_ERROR_MESSAGE_FORMAT = 'Artman user YAML %s is invalid.'


ARTIFACT_MAPPING = {
    'java_proto': {'language': 'JAVA', 'type': 'PROTOBUF'},
    'java_grpc': {'language': 'JAVA', 'type': 'GRPC'},
    'ruby_grpc': {'language': 'RUBY', 'type': 'GRPC'},
    'python_grpc': {'language': 'PYTHON', 'type': 'GRPC'},
    'java_gapic': {'language': 'JAVA', 'type': 'GAPIC'},
    'java_discogapic': {'language': 'JAVA', 'type': 'DISCOGAPIC'},
    'java_gapic_only': {'language': 'JAVA', 'type': 'GAPIC_ONLY'},
    'python_gapic': {'language': 'PYTHON', 'type': 'GAPIC'},
    'nodejs_gapic': {'language': 'NODEJS', 'type': 'GAPIC'},
    'php_gapic': {'language': 'PHP', 'type': 'GAPIC'},
    'go_gapic': {'language': 'GO', 'type': 'GAPIC'},
    'ruby_gapic': {'language': 'RUBY', 'type': 'GAPIC'},
    'csharp_gapic': {'language': 'CSHARP', 'type': 'GAPIC'},
    'gapic_config': {'type': 'GAPIC_CONFIG'},
    'discogapic_config': {'type': 'DISCOGAPIC_CONFIG'},
    'php_grpc': {'language': 'JAVA', 'type': 'GRPC'},
}


def load_artifact_config(artman_config_path, artifact_name, aspect=None):
    artman_config = _read_artman_config(artman_config_path)
    artifact_config = Artifact()
    artifact_config.CopyFrom(artman_config.common)

    if artifact_name not in ARTIFACT_MAPPING:
        raise ValueError(
            'Artifact type `%s` not supported. Valid values are %s'
            % (artifact_name, ARTIFACT_MAPPING.keys()))
    loaded_artifact = Artifact(**ARTIFACT_MAPPING[artifact_name])

    for artifact in artman_config.artifacts:
        if artifact.name == artifact_name:
            loaded_artifact.MergeFrom(artifact)
            break

    if aspect:
        loaded_artifact.aspect = Artifact.Aspect.Value(aspect)

    artifact_config.MergeFrom(loaded_artifact)
    _validate_artifact_config(artifact_config)
    return _normalize_artifact_config(
        artifact_config, artman_config_path)


def read_user_config(artman_user_config_path):
    """Parse and return artman config"""
    config_pb = UserConfig()
    artman_user_config_path = os.path.expanduser(artman_user_config_path)
    if not os.path.isfile(artman_user_config_path):
      logger.warn(
          'No artman user config defined. Use the default one for this '
          'execution. Run `configure-artman` to set up user config.')
      return config_pb

    try:
        with io.open(artman_user_config_path, 'r', encoding='UTF-8') as f:
            # Convert yaml into json file as protobuf python load support
            # parsing of protobuf in json or text format, not yaml.
            json_string = json.dumps(yaml.load(f, Loader=yaml.FullLoader))
        json_format.Parse(json_string, config_pb)
    except (json_format.ParseError, yaml.parser.ParserError):
        logger.error(INVALID_USER_CONFIG_ERROR_MESSAGE_FORMAT % artman_user_config_path)
        raise

    return config_pb

def _read_artman_config(artman_yaml_path):
    """Parse and return artman config after validation and normalization."""
    artman_config = _parse(artman_yaml_path)
    validation_result = _validate_artman_config(artman_config)
    if validation_result:
        raise ValueError(validation_result)
    else:
        return artman_config


def _parse(artman_yaml_path):
    """Parse artman yaml config into corresponding protobuf."""
    if not os.path.exists(artman_yaml_path):
        raise ValueError(CONFIG_NOT_FOUND_ERROR_MESSAGE_FORMAT % artman_yaml_path)

    try:
        with io.open(artman_yaml_path, 'r', encoding='UTF-8') as f:
            # Convert yaml into json file as protobuf python load support paring of
            # protobuf in json or text format, not yaml.
            artman_config_json_string = json.dumps(yaml.load(f, Loader=yaml.FullLoader))
        config_pb = Config()
        json_format.Parse(artman_config_json_string, config_pb)
    except (json_format.ParseError, yaml.parser.ParserError):
        logger.error(INVALID_CONFIG_ERROR_MESSAGE_FORMAT % artman_yaml_path)
        raise

    return config_pb


def _validate_artman_config(config_pb):
    """ Validate the parsed config_pb.

    Currently, it checkes the uniqueness of the artifact name, and uniqieuness
    of publishing target name in each artifact. Plus, it makes sure the
    specified configuration or other input file or folder exists.
    """
    artifacts = set()
    for artifact in config_pb.artifacts:
        if artifact.name in artifacts:
            return ('artifact `%s` has been configured twice, please rename.' %
                    artifact.name)
        artifacts.add(artifact.name)

    return None


def _validate_artifact_config(artifact_config):
    if (artifact_config.language == Artifact.NODEJS and
            artifact_config.type == Artifact.GRPC):
        raise ValueError('GRPC artifact type is invalid for NodeJS.')


def _normalize_artifact_config(artifact_config, artman_config_path):
    """Normalize the config protobuf based on flags passed from command line.

    Note: we are not normalizing output folders because they are no longer
    configurable, and the output folder calculation now happens in converter.
    Once the individual GAPIC output folder becomes configurable, that folder
    name calculation logic should be moved from converter into this method.
    """
    # Normalize the input file or folder by concatenating relative path with
    # the folder of artman config yaml.
    if artifact_config.service_yaml:
        artifact_config.service_yaml = _normalize_path(
            artifact_config.service_yaml, artman_config_path, 'service_yaml')

    if artifact_config.gapic_yaml:
        artifact_config.gapic_yaml = _normalize_path(
            artifact_config.gapic_yaml, artman_config_path, 'gapic_yaml')

    if artifact_config.samples:
        artifact_config.samples = _normalize_path(
            artifact_config.samples, artman_config_path, 'samples')

    if artifact_config.grpc_service_config:
        artifact_config.grpc_service_config = _normalize_path(
            artifact_config.grpc_service_config, artman_config_path, 'grpc_service_config')

    normalized_src_proto_paths = []
    for src_proto_path in artifact_config.src_proto_paths:
        if src_proto_path.startswith('-'):
            # Retain the exclusion mark "-" as the prefix of the normalized path
            normalized_src_proto_paths.append(
                '-%s' % _normalize_path(src_proto_path[1:],
                                        artman_config_path,
                                        'src_proto_paths'))
        else:
            normalized_src_proto_paths.append(
                _normalize_path(src_proto_path,
                                artman_config_path,
                                'src_proto_paths'))
    artifact_config.src_proto_paths[:] = normalized_src_proto_paths

    return artifact_config


def _normalize_path(path, artman_config_path, field):
    if path.startswith('..') or '/..' in path:
        raise ValueError(
            '".." is disallowed in `%s` field of `%s`. Please use either a '
            'path relative to `%s` (preferred), or an absolute path.'
            % (field, artman_config_path,
               os.path.dirname(artman_config_path)))
    if os.path.isabs(path):
        return path
    else:
        return os.path.join(os.path.dirname(artman_config_path), path)



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for pipeline."""

from taskflow.flow import Flow
from taskflow.patterns import linear_flow


class PipelineBase(object):
    """Base class of pipeline.

    Subclasses must implement validate_kwargs and do_build_flow methods.
    """

    def __init__(self, **kwargs):
        self._kwargs = kwargs
        self._flow = self.build_flow(**kwargs)

    def build_flow(self, **kwargs):
        """Make the task flow based on kwargs."""
        self.validate_kwargs(**kwargs)

        # Call do_make_flow implemented by subclass to make the flow.
        flow = self.do_build_flow(**kwargs)
        if not isinstance(flow, Flow):
            raise TypeError('Return type must be taskflow.flow.Flow.')

        return flow

    def validate_kwargs(self, **kwargs):
        """Abstract method, subclass must implement this method to validate
        kwargs.
        """
        raise NotImplementedError('Subclass must implement abstract method')

    def do_build_flow(self, **kwargs):
        """Abstract method, subclass must implment this method and return a task
        instance.
        """
        raise NotImplementedError('Subclass must implement abstract method')

    @property
    def flow(self):
        return self._flow

    @property
    def name(self):
        return self.__class__.__name__

    @property
    def kwargs(self):
        return self._kwargs


class EmptyPipeline(PipelineBase):

    def do_build_flow(self, **kwargs):
        flow = linear_flow.Flow('empty-pipeline')
        return flow

    def validate_kwargs(self, **kwargs):
        pass



# Copyright 2018 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tasks that work directly with a descriptor set"""

import os
import re

import pypandoc
from artman.tasks import task_base

from google.protobuf import descriptor_pb2 as desc


class PythonDocsConvertionTask(task_base.TaskBase):
    """Converts proto comments to restructuredtext format.

    Proto comments are expected to be in markdown format, and to possibly
    contain links to other protobuf types and relative URLs that will not
    resolve to correct documentation using standard tools.

    This task performs the following transformations on the documentation
    in the descriptor set:
    - Replace proto links with literals (e.g. [Foo][bar.baz.Foo] -> `Foo`)
    - Resolve relative URLs to https://cloud.google.com
    - Run pandoc to convert from markdown to restructuredtext"""
    default_provides = 'descriptor_set'

    def execute(self, descriptor_set):
        desc_set = desc.FileDescriptorSet()
        with open(descriptor_set, 'rb') as f:
            desc_set.ParseFromString(f.read())

        for file_descriptor_proto in desc_set.file:
            if not file_descriptor_proto.source_code_info:
                continue

            for location in file_descriptor_proto.source_code_info.location:
                location.leading_comments = md2rst(location.leading_comments)
                location.trailing_comments = md2rst(location.trailing_comments)
                detached = [md2rst(c) for c in location.leading_detached_comments]
                del location.leading_detached_comments[:]
                location.leading_detached_comments.extend(detached)

        desc_file, desc_ext = os.path.splitext(descriptor_set)
        new_descriptor_set = desc_file + '_updated_py_docs' + desc_ext

        with open(new_descriptor_set, mode='wb') as f:
            f.write(desc_set.SerializeToString())

        return new_descriptor_set


_proto_link_re = re.compile(
    r'(\[(?P<text>[^\]]+)\]\[(?P<uri>[A-Za-z_][A-Za-z_.0-9]*)?\])')
_relative_link_re = re.compile(
    r'(\[(?P<text>[^\]]+)\]\((?P<uri>/[^\)]+)\))')
_new_lines = re.compile(r'(?P<newlines>(\r?\n)+)(?P<followup>[^\r\n])')


def _replace(comment, pattern, repl_fn):
    index = 0
    strs = []
    for m in pattern.finditer(comment):
        strs.append(comment[index:m.start()])
        strs.append(repl_fn(m))
        index = m.end()
    strs.append(comment[index:])
    return ''.join(strs)


def _replace_proto_link(comment):
    def _format(m):
        return "`{}`".format(m.group('text'))
    return _replace(comment, _proto_link_re, _format)


def _replace_relative_link(comment):
    def _format(m):
        return "[{}](https://cloud.google.com{})".format(m.group('text'),
                                                         m.group('uri'))
    return _replace(comment, _relative_link_re, _format)


def _insert_spaces(comment):
    def _format(m):
        return "{} {}".format(m.group('newlines'), m.group('followup'))
    # After replacement, add a leading space
    return " " + _replace(comment, _new_lines, _format)


def _add_space(comment):
    def _format(m):
        return "[{}](https://cloud.google.com{})".format(m.group('text'),
                                                         m.group('uri'))
    return _replace(comment, _relative_link_re, _format)


def md2rst(comment):
    """Convert a comment from protobuf markdown to restructuredtext.

    This method:
    - Replaces proto links with literals (e.g. [Foo][bar.baz.Foo] -> `Foo`)
    - Resolves relative URLs to https://cloud.google.com
    - Runs pandoc to convert from markdown to restructuredtext
    """
    comment = _replace_proto_link(comment)
    comment = _replace_relative_link(comment)
    # Calling pypandoc.convert_text is slow, so we try to avoid it if there are
    # no special characters in the markdown.
    if any([i in comment for i in '`[]*_']):
        comment = pypandoc.convert_text(comment, 'rst', format='commonmark')
        # Comments are now valid restructuredtext, but there is a problem. They
        # are being inserted back into a descriptor set, and there is an
        # expectation that each line of a comment will begin with a space, to
        # separate it from the '//' that begins the comment. You would think
        # that we could ignore this detail, but it will cause formatting
        # problems down the line in gapic-generator because parsing code will
        # try to remove the leading space, affecting the indentation of lines
        # that actually do begin with a space, so we insert the additional
        # space now. Comments that are not processed by pypandoc will already
        # have a leading space, so should not be changed.
        comment = _insert_spaces(comment)
    return comment


_DESC_TASK_DICT = {
    'python': PythonDocsConvertionTask,
}


class EmptyDescriptorTask(task_base.EmptyTask):
    """An empty descriptor task.

     The taskflow library will produce a duplicate task error if multiple
     EmptyTask instances are added to a pipeline, so we create a unique
     subclass."""
    pass


def get_descriptor_set_task(language):
    return _DESC_TASK_DICT.get(language, EmptyDescriptorTask)



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pytest
import unittest
from ruamel import yaml

from artman.tasks import package_metadata_tasks


class PackageMetadataConfigTest(unittest.TestCase):
    # Print full diff when comparing actual/expected dict output
    maxDiff = None

    @pytest.fixture(autouse=True)
    def outdir(self, tmpdir):
        self.output_dir = tmpdir.mkdir('out')

    def test_package_metadata_config_gen_task(self):
        task = package_metadata_tasks.PackageMetadataConfigGenTask()
        repo_root = os.path.abspath('.')

        task.execute(
            api_name='fake',
            api_version='v1',
            language='python',
            root_dir='%s/googleapis' % repo_root,
            organization_name='google-cloud',
            output_dir=str(self.output_dir),
            proto_deps=['googleapis-common-protos'],
            artifact_type='GAPIC',
            src_proto_path=['path/to/protos'],
            release_level='beta'
        )
        actual_file = os.path.join(str(self.output_dir),
                                   'python_google-cloud-fake-v1_package2.yaml')
        expected_file = 'test/testdata/python_google-cloud-fake-v1_package2.yaml'
        with open(actual_file) as f:
            actual = yaml.safe_load(f)
        with open(expected_file) as f:
            expected = yaml.safe_load(f)
        # Don't compare files directly because yaml doesn't preserve ordering
        try:
            self.assertDictEqual(actual, expected)
        except:
            print("comparison failure: actual = " + actual_file + ", expected = " + expected_file)
            raise
